import re

from .exceptions import InstructionError
from .memory import Memory
from .mnemonic import MNEMONIC


def format_line(addr, text, code):
    items = text.split(" ", maxsplit=2)
    return f"{items[0]:8}{' '.join(items[1:]):32};[{addr:04x}] " + " ".join(
        [f"{c:02x}" for c in code]
    )


def disasm_line(mem):
    addr = mem.addr
    op = mem.next_byte()
    if op is None:
        return -1, ""
    try:
        func = MNEMONIC.get(op)
        text = func(op, mem)
        line = format_line(addr, text, mem[addr : mem.addr])
        return addr, line
    except InstructionError as e:
        raise InstructionError(f"{e} at {addr:04x}")


def disasm_nlines(mem, addr, max_line):
    if not mem.addr_in(addr):
        print(f"start address is out of range {addr:04x}")
        exit()

    mem.addr = addr
    op = mem.next_byte()
    lines = {}
    count = 0
    while op is not None:
        func = MNEMONIC.get(op)
        try:
            text = func(op, mem)
            line = format_line(addr, text, mem[addr : mem.addr])
            print(" " * 16 + line)
            lines[addr] = line
        except Exception as e:
            print(e, f"at {addr:04x}")
            return lines

        lines[addr] = format_line(addr, text, mem[addr : mem.addr])
        addr = mem.addr
        op = mem.next_byte()
        count += 1
        if count >= max_line:
            break
    return lines


def get_branchs(lines):
    branches = {}
    for addr, text in lines.items():
        m = re.search(r"(?:JP|CALL|JR).*?\$([0-9A-F]{4})", text, flags=re.IGNORECASE)
        if m is None:
            continue
        branches[addr] = int(m.group(1), base=16)
    return branches


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        exit()
    mem = Memory(open(sys.argv[1], "rb").read())
    disasm_nlines(mem)



from .mnemonic_defs import REG8, REG16_SP


def in_r(op, _):
    r = (op >> 3) & 7
    if r == 6:
        return "IN (C)"
    else:
        return f"IN {REG8[r]},(C)"


def out_r(op, _):
    r = (op >> 3) & 7
    if r == 6:
        return "OUT (C),0"
    else:
        return f"OUT (C),{REG8[r]}"


def sbc_hl(op, _):
    rr = (op >> 4) & 3
    return f"SBC HL,{REG16_SP[rr]}"


def adc_hl(op, _):
    rr = (op >> 4) & 3
    return f"ADC HL,{REG16_SP[rr]}"


def ld_mem_rr(op, mem):
    rr = (op >> 4) & 3
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD (${n2:02X}{n1:02X}),{REG16_SP[rr]}"


def ld_rr_mem(op, mem):
    rr = (op >> 4) & 3
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD {REG16_SP[rr]},(${n2:02X}{n1:02X})"


MNEMONIC_ED = {
    0x40: in_r,
    0x41: out_r,
    0x42: sbc_hl,
    0x4A: adc_hl,
    0x43: ld_mem_rr,
    0x53: ld_mem_rr,
    0x63: ld_mem_rr,
    0x73: ld_mem_rr,
    0x4B: ld_rr_mem,
    0x6B: ld_rr_mem,
    0x5B: ld_rr_mem,
    0x7B: ld_rr_mem,
    0x47: lambda *_: "LD I,A",
    0x57: lambda *_: "LD A,I",
    0x4F: lambda *_: "LD R,A",
    0x5F: lambda *_: "LD A,R",
    0xA0: lambda *_: "LDI",
    0xA8: lambda *_: "LDD",
    0xB0: lambda *_: "LDIR",
    0xB8: lambda *_: "LDDR",
    0xA1: lambda *_: "CPI",
    0xA9: lambda *_: "CPD",
    0xB1: lambda *_: "CPIR",
    0xB9: lambda *_: "CPDR",
    0xA2: lambda *_: "INI",
    0xAA: lambda *_: "IND",
    0xB2: lambda *_: "INIR",
    0xBA: lambda *_: "INDR",
    0xA3: lambda *_: "OUTI",
    0xAB: lambda *_: "OUTD",
    0xB3: lambda *_: "OTIR",
    0xBB: lambda *_: "OTDR",
    0x44: lambda *_: "NEG",
    0x45: lambda *_: "RETN",
    0x4D: lambda *_: "RETI",
    0x46: lambda *_: "IM 0",
    0x56: lambda *_: "IM 1",
    0x5E: lambda *_: "IM 2",
    0x67: lambda *_: "RRD",
    0x6F: lambda *_: "RLD",
}


def init_instruction_dict():
    for n in range(1, 8):
        MNEMONIC_ED[0x40 + n * 8] = MNEMONIC_ED[0x40]  # IN r,(C)
        MNEMONIC_ED[0x41 + n * 8] = MNEMONIC_ED[0x41]  # OUT (C),r

    for n in range(1, len(REG16_SP)):
        MNEMONIC_ED[0x42 + n * 0x10] = MNEMONIC_ED[0x42]  # SBC HL,rr
        MNEMONIC_ED[0x4A + n * 0x10] = MNEMONIC_ED[0x4A]  # ADC HL,rr


init_instruction_dict()



REG8 = ["B", "C", "D", "E", "H", "L", "(HL)", "A"]
REG16_SP = ["BC", "DE", "HL", "SP"]
REG16_AF = ["BC", "DE", "HL", "AF"]

ARITHMETIC = ["ADD A,", "ADC A,", "SUB ", "SBC A,", "AND ", "XOR ", "OR ", "CP "]
ROTATE_SHIFT = ["RLCA", "RRCA", "RLA", "RRA", "DAA", "CPL", "SCF", "CCF"]

ROTATE_SHIFT_R = ["RLC", "RRC", "RL", "RR", "SLA", "SRA", None, "SRL"]
BIT_OP = [None, "BIT", "RES", "SET"]

CC = ["NZ", "Z", "NC", "C", "PO", "PE", "P", "M"]


def uint8_to_int8(value):
    if value <= 127:
        return value
    else:
        return value - 256



import re
from collections import defaultdict
from dataclasses import dataclass

from .disasm import disasm_line

debug = False


@dataclass
class Label:
    addr: int
    label_type: set
    used_addr: set
    processed: bool
    name_cache: str = ""

    @property
    def name(self):
        if self.name_cache != "":
            return self.name_cache

        base = "".join(sorted(self.label_type))
        self.name_cache = f"{base}_{self.addr:04X}"
        return self.name_cache

    def check_external(self, mem):
        if mem.addr_in(self.addr):
            return
        if not self.name.startswith("EX_"):
            self.name_cache = "EX_" + self.name_cache


def add_branch_label(labels, addr, line):
    LABEL_TYPE = {"CALL": "CD", "JR": "JR", "JP": "JP", "DJNZ": "JR"}
    m = re.search(
        r"^\s*(JP|JR|DJNZ|CALL)\s+.*?\$([0-9a-f]{4})", line, flags=re.IGNORECASE
    )
    if m is None:
        return
    target = int(m.group(2), base=16)
    label = labels.setdefault(target, Label(target, set(), set(), False))
    label.label_type.add(LABEL_TYPE[m.group(1)])
    label.used_addr.add(addr)


def add_data_label(labels, addr, line):
    m = re.search(r"\(\$([0-9a-f]{4})\)", line, flags=re.IGNORECASE)
    if m is None:
        return
    target = int(m.group(1), base=16)
    label = labels.setdefault(target, Label(target, set(), set(), False))
    label.label_type.add("DT")
    label.used_addr.add(addr)


STOP = set(["RET", "RETI", "RETN", "HALT"])


def should_pause(line):
    code = line.split(";")[0].strip()
    if code in STOP:
        return True
    m = re.search(r"^\s*(JP|JR)\s+[\$(]", line)
    return m is not None


def in_range(ranges, addr):
    return any(addr in r for r in ranges)


def replace_branch_addr_ref(labels, lines):
    for target, label in labels.items():
        # if target not in lines:
        #     continue
        for addr in label.used_addr:
            lines[addr] = lines[addr].replace(f"${target:04X}", label.name)


def addr_label(addr, branch_labels, data_labels):
    if addr in branch_labels:
        return branch_labels[addr].name
    if addr in data_labels:
        return data_labels[addr].name


def bytes2ascii(bstr):
    block = bytearray(bstr)
    for n, b in enumerate(block):
        if b < 0x20 or b >= 0x7E:
            block[n] = ord(".")
    return block.decode("ascii")


def bytes2string(bstr):
    ret = ""
    in_str = False
    if 0x20 <= bstr[0] <= 0x7E:
        in_str = True
        ret += '"'
    for b in bstr:
        if 0x20 <= b <= 0x7E and in_str:
            c = chr(b) if b != 0x22 else r"\""
            ret += c
        elif 0x20 <= b <= 0x7E:
            in_str = True
            ret += '"'
            c = chr(b) if b != 0x22 else r"\""
            ret += c
        elif in_str:
            in_str = False
            ret += f'",${b:02X},'
        else:
            ret += f"${b:02X},"
    if in_str:
        return ret + '"'
    else:
        return ret[:-1]


def scan_str_ref(lines, branch_labels):
    target_addrs = set(
        addr for addr, lbl in branch_labels.items() if "ST" in lbl.label_type
    )
    for addr, line in lines.items():
        m = re.search(r"(\(?\$[0-9a-f]{4}\)?)", line, flags=re.IGNORECASE)
        if m is None:
            continue
        if m.group(1)[0] == "$":
            ref_addr = int(m.group(1)[1:], base=16)
        else:
            ref_addr = int(m.group(1)[2:6], base=16)
        if ref_addr not in target_addrs:
            continue
        branch_labels[ref_addr].used_addr.add(addr)


def merge_ranges(ranges):
    ranges.sort(key=lambda r: r.start)

    merged = True
    ofs = 0
    while merged:
        merged = False
        for n in range(ofs, len(ranges) - 1):
            if ranges[n].stop < ranges[n + 1].start:
                ofs += 1
                continue
            start = min(ranges[n].start, ranges[n + 1].start)
            stop = max(ranges[n].stop, ranges[n + 1].stop)
            ranges[n : n + 2] = [range(start, stop)]
            merged = True
            break


def creat_data_ranges(code_ranges, min_addr, max_addr, label_addrs):
    data_ranges = []

    min_start = min(r.start for r in code_ranges)
    if min_start < min_addr:
        data_ranges.append(range(min_addr, min_start - 1))
    max_stop = max(r.stop for r in code_ranges)
    if max_stop <= max_addr:
        data_ranges.append(range(max_stop, max_addr + 1))

    for n, rng in enumerate(code_ranges[:-1]):
        data_ranges.append(range(rng.stop, code_ranges[n + 1].start))

    if debug:
        breakpoint()

    ret_ranges = []
    addr_set = set(label_addrs)
    for rng in data_ranges:
        if not addr_set:
            ret_ranges.append(rng)
            continue
        addrs = sorted(a for a in addr_set if a in rng)
        if not addrs:
            ret_ranges.append(rng)
            continue
        start = rng.start
        for addr in addrs:
            ret_ranges.append(range(start, addr))
            start = addr
        ret_ranges.append(range(start, rng.stop))
        addr_set -= set(addrs)

    return ret_ranges


def create_db_lines(lines, data_ranges, mem):
    for rng in data_ranges:
        lines.update(set_db_line(mem, rng))


def set_db_line(mem, rng):
    lines = {}
    for addr in range(rng.start, rng.stop, 8):
        block = bytearray(mem[addr : min(addr + 8, rng.stop)])
        line = "DB      " + ",".join(f"${b:02X}" for b in block)
        line += "    " * (8 - len(block))
        for n, b in enumerate(block):
            if b < 0x20 or b >= 0x7E:
                block[n] = ord(".")
        line += f"   ;[{addr:04x}] " + block.decode("ascii")
        lines[addr] = line
    return lines


def define_equ(mem, line_addrs, *group):
    for labels in group:
        for addr in sorted(labels.keys()):
            if not mem.addr_in(addr):
                print(f"{labels[addr].name:16}EQU     ${labels[addr].addr:04x}")
                continue
            if addr not in line_addrs:
                print(
                    f"{labels[addr].name:16}EQU     ${labels[addr].addr:04x} ; within CODE"
                )
                continue
    print("")


def disasm_eagerly(args, mem):
    global debug
    if args.debug:
        debug = True

    ranges = []
    lines = {}
    branch_labels = defaultdict(dict)
    data_labels = defaultdict(dict)

    # --code
    for rng in args.code:
        ranges.append(rng)
        mem.addr = rng.start
        print(f"; start: {mem.start:04x}")
        while mem.addr < rng.stop:
            try:
                addr, line = disasm_line(mem)
                if line == "":
                    break
                lines[addr] = line
                add_branch_label(branch_labels, addr, line)
                add_data_label(data_labels, addr, line)
            except Exception as e:
                print(e)
                # exit()

    # --string
    for rng in args.string:
        ranges.append(rng)
        addr = rng.start
        text = bytes2string(mem[addr : rng.stop])
        line = f'DB    {text} ;[{addr:04x}] {" ".join(f"{b:02x}" for b in mem[addr:rng.stop])}'
        lines[addr] = line
        branch_labels[addr] = Label(addr, "ST", set(), True)

    # --addr
    addrs = args.addr
    if not lines and not addrs:
        # no --code, no --addr
        addrs = [mem.min_addr]

    for start_addr in addrs:
        if start_addr in lines:
            continue
        mem.addr = start_addr
        while True:
            try:
                addr, line = disasm_line(mem)
                if line == "":
                    break
                lines[addr] = line
                add_branch_label(branch_labels, addr, line)
                add_data_label(data_labels, addr, line)
            except Exception as e:
                print(e)
                exit()
            if should_pause(line):
                break
        ranges.append(range(start_addr, mem.addr))

    # branch addresses
    while True:
        branches = sorted(a for a, lbl in branch_labels.items() if not lbl.processed)
        if not branches:
            break
        for start_addr in branches:
            if start_addr in lines or not mem.addr_in(start_addr):
                branch_labels[start_addr].processed = True
                continue

            mem.addr = start_addr
            while True:
                addr, line = disasm_line(mem)
                if line == "":
                    break
                lines[addr] = line
                add_branch_label(branch_labels, addr, line)
                add_data_label(data_labels, addr, line)
                if should_pause(line):
                    ranges.append(range(start_addr, mem.addr))
                    break

    scan_str_ref(lines, branch_labels)
    merge_ranges(ranges)

    # DB
    data_ranges = creat_data_ranges(
        ranges, mem.min_addr, mem.max_addr, sorted(data_labels.keys())
    )
    create_db_lines(lines, data_ranges, mem)

    # add EX_
    for labels in [branch_labels, data_labels]:
        for label in labels.values():
            label.check_external(mem)

    # add labels
    replace_branch_addr_ref(branch_labels, lines)
    replace_branch_addr_ref(data_labels, lines)

    # define external and in-code label with  EQU
    define_equ(mem, lines.keys(), branch_labels, data_labels)

    addrs = sorted(lines.keys())
    # ORG
    print(" " * 16 + f"ORG     ${addrs[0]:04X}\n")

    for addr in addrs:
        label = addr_label(addr, branch_labels, data_labels)
        if label is None:
            print(" " * 16, end="")
        elif len(label) < 15:
            label += ":"
            print(f"\n{label:16}", end="")
        else:
            print(f"\n{label}:")
        cols = lines[addr].split(";")
        print(f"{cols[0].strip():40}; {cols[1].strip()}")

    # data or code
    output_information(mem, branch_labels, data_labels, data_ranges)


def output_information(mem, branch_labels, data_labels, data_ranges):
    def print_xref(name, refs):
        ADDR_PER_LINE = 10
        for index in range(0, len(refs), ADDR_PER_LINE):
            print(f"; {name:16}", end="")
            print(
                " ".join(
                    f"${x:04x}"
                    for x in refs[index : min(index + ADDR_PER_LINE, len(refs))]
                )
            )
            name = ""

    print("\n; XREF information")
    for addr in sorted(list(branch_labels.keys() | data_labels.keys())):
        if addr in branch_labels:
            print_xref(branch_labels[addr].name, sorted(branch_labels[addr].used_addr))
        if addr in data_labels:
            print_xref(data_labels[addr].name, sorted(data_labels[addr].used_addr))

    print("\n; DATA summary")
    for rng in data_ranges:
        decoded = bytes2ascii(mem[rng.start : rng.stop])
        print(f"; ${rng.start:04x}-${rng.stop - 1:04x}, [${len(rng):4x}] ", end="")
        print(decoded[:48])



import argparse
import re
import sys
from pathlib import Path

from .disasm import disasm_nlines
from .eager import disasm_eagerly
from .loader import load

DEFAULT_MAX_LINES = 32


def parse_addr(arg):
    arg = arg.upper()
    if arg.startswith("$"):
        arg = arg[1:]
    if arg.endswith("H"):
        arg = arg[:-1]
    try:
        return int(arg, base=16)
    except:  # noqa:E722
        raise argparse.ArgumentTypeError(f"invalid address: {arg}")


def parse_range(arg):
    addrs = arg.split("-")
    if len(addrs) != 2:
        raise argparse.ArgumentTypeError
    start = parse_addr(addrs[0])
    stop = parse_addr(addrs[1]) + 1
    if start >= stop:
        raise argparse.ArgumentTypeError
    return range(start, stop)


def parse_option_file(file):
    if not Path(file).exists():
        raise argparse.ArgumentTypeError(f"'{file}' not found")

    options = []
    with open(file) as fp:
        for line in fp:
            line = re.sub(r"[#;].*$", "", line).strip()
            if not line:
                continue
            cols = line.split()
            if cols[0] not in {
                "--eager",
                "-e",
                "--code",
                "-c",
                "--string",
                "-s",
                "--addr",
                "-a",
                "--max-lines",
                "-m",
                "--offset",
                "-o",
            }:
                raise argparse.ArgumentTypeError(f"unrecognized option {cols[0]}")
            options.extend(cols)
        return options


def build_parser():
    parser = argparse.ArgumentParser(prog="yad80")
    parser.add_argument("--version", "-v", action="version", version="%(prog)s 0.1.4")
    parser.add_argument(
        "--option", nargs=1, type=parse_option_file, default=[], help="option file"
    )
    parser.add_argument(
        "--code",
        "-c",
        action="extend",
        nargs="*",
        type=parse_range,
        default=[],
        metavar="RANGE",
        help="address range(a1-a2) as code. a2 is an inclusive address",
    )
    parser.add_argument(
        "--string",
        "-s",
        action="extend",
        nargs="*",
        type=parse_range,
        default=[],
        metavar="RANGE",
        help="address range(a1-a2) as string. a2 is an inclusive address",
    )
    parser.add_argument(
        "--addr",
        "-a",
        action="extend",
        nargs="*",
        type=parse_addr,
        default=[],
        metavar="ADDR",
        help="address to disasm",
    )
    parser.add_argument(
        "--eager", "-e", action="store_true", help="disasm eagerly(default false)"
    )
    parser.add_argument(
        "--debug", action="store_true", default=False, help="debug flag(dev use)"
    )
    parser.add_argument(
        "--max-lines",
        "-m",
        type=int,
        default=DEFAULT_MAX_LINES,
        metavar="N",
        help=f"max lines to output(default {DEFAULT_MAX_LINES})",
    )
    parser.add_argument(
        "--offset", "-o", type=parse_addr, default=0, help="address offset for binary file"
    )
    parser.add_argument("FILE", help="file to disasm")

    return parser


def parse_args(args):
    parser = build_parser()
    parsed = parser.parse_args(args)
    if not parsed.option:
        return parsed

    base = parser.parse_args(parsed.option[0] + ["--", parsed.FILE])
    base.code.extend(parsed.code)
    base.string.extend(parsed.string)
    base.addr.extend(parsed.addr)
    base.eager = base.eager or parsed.eager
    if parsed.max_lines != DEFAULT_MAX_LINES:
        base.max_lines = parsed.max_lines
    if parsed.offset != 0:
        base.offset = parsed.offset

    return base


def cli_main(argv):
    args = parse_args(argv)

    mem = load(args.FILE, args.offset)

    if args.eager:
        disasm_eagerly(args, mem)
        return

    if not args.addr:
        start_addr = mem.start
    elif len(args.addr) == 1:
        start_addr = args.addr[0]
    else:
        print(f"mulitple address {args.addr} specified")
        return

    disasm_nlines(mem, start_addr, args.max_lines)


def main():
    cli_main(sys.argv[1:])
